#include "SimpleAnalysis/AnalysisClass.h"

DefineAnalysis(ThreeBjetsEWK2016)


void ThreeBjetsEWK2016::Init()
{

  // All variables -- don't care how these are affected by TRF/non-TRF
  addHistogram("met",20,0,2000);
  addHistogram("met_sig",20,0,50);
  addHistogram("mjsum",20,0,2000);
  addHistogram("mTb_min",20,0,800);
  addHistogram("mT", 20, 0, 500);
  addHistogram("dphiMin4", 50, 0, 5);
  addHistogram("dphi1jet", 50, 0, 5);
  addHistogram("n_jets", 20, -0.5, 19.5);
  addHistogram("n_bjets",20, -0.5, 19.5);
  addHistogram("n_electrons", 20, -0.5, 19.5);
  addHistogram("n_muons", 20, -0.5, 19.5);
  addHistogram("n_leptons", 20, -0.5, 19.5);
  addHistogram("GenMET",40,0,2000);
  addHistogram("GenHT",40,0,2000);
  addHistogram("mc_weight", 1, 0, 1);

  // Variables for non-TRF
  addHistogram("meff_4bj",20,0,2000);
  addHistogram("m1", 40, 0, 1000);
  addHistogram("m2", 40, 0, 1000);
  addHistogram("DeltaR_max", 50, 0, 10);
  
  // Variables for TRF
  addHistogram("meff_4bj_ex3b",20,0,2000);
  addHistogram("meff_4bj_in3b",20,0,2000);
  addHistogram("meff_4bj_in4b",20,0,2000);
  addHistogram("meff_4bj_in2b",20,0,2000);
  addHistogram("m1_ex3b", 40, 0, 1000);
  addHistogram("m2_ex3b", 40, 0, 1000);
  addHistogram("m1_in3b", 40, 0, 1000);
  addHistogram("m2_in3b", 40, 0, 1000);
  addHistogram("m1_in4b", 40, 0, 1000);
  addHistogram("m2_in4b", 40, 0, 1000);
  addHistogram("m1_in2b", 40, 0, 1000);
  addHistogram("m2_in2b", 40, 0, 1000);
  addHistogram("DeltaR_max_ex3b", 50, 0, 10);
  addHistogram("DeltaR_max_in3b", 50, 0, 10);
  addHistogram("DeltaR_max_in4b", 50, 0, 10);
  addHistogram("DeltaR_max_in2b", 50, 0, 10);
  
// This analysis requires an additional BTaggingTruthTagging tool
#ifdef ROOTCORE_PACKAGE_BTaggingTruthTagging
#pragma message "Compiling BTaggingTruthTagging for TRF usage"
  m_btt = new BTaggingTruthTaggingTool("MyBTaggingTruthTaggingTool");
  StatusCode code = m_btt->setProperty("TaggerName",          "MV2c10");
  if (code != StatusCode::SUCCESS) throw std::runtime_error("error setting BTaggingTruthTaggingTool TaggerName property");
  code = m_btt->setProperty("OperatingPoint", "FixedCutBEff_77");
  if (code != StatusCode::SUCCESS) throw std::runtime_error("error setting BTaggingTruthTaggingTool OperatingPoint property");
  code = m_btt->setProperty("JetAuthor", "AntiKt4EMTopoJets");
  if (code != StatusCode::SUCCESS) throw std::runtime_error("error setting BTaggingTruthTaggingTool JetAuthor property");
  code = m_btt->setProperty("ScaleFactorFileName", "xAODBTaggingEfficiency/13TeV/2016-20_7-13TeV-MC15-CDI-2017-01-31_v1.root");
  if (code != StatusCode::SUCCESS) throw std::runtime_error("error setting BTaggingTruthTaggingTool ScaleFactorFileName property");
  // call initialize() function
  code = m_btt->initialize();
  if (code != StatusCode::SUCCESS) throw std::runtime_error("error initializing the BTaggingTruthTaggingTool");
  std::cout << "Initialized BTaggingTruthTagging tool" << std::endl;
#endif

}

static int jetFlavor(AnalysisObject &jet) {
  if (jet.pass(TrueBJet)) return 5;
  if (jet.pass(TrueCJet)) return 4;
  if (jet.pass(TrueTau)) return 15;
  return 0;
}

void ThreeBjetsEWK2016::ProcessEvent(AnalysisEvent *event)
{

  float gen_met      = event->getGenMET();
  float gen_ht       = event->getGenHT();
  int channel_number = event->getMCNumber();
  fill("mc_weight", 0.5, event->getMCWeights()[0]);


  // MET-filtered and inclusive MC used for electroweak Multi-b
  // For ttbar
  if(channel_number==410001 && gen_met>200) return;
  if(channel_number==410002 && gen_met>200) return;
  if(channel_number==410225 && gen_ht>600) return; // This sample is Ht-filtered 
  if(channel_number==410004 && gen_met>200) return;
  if(channel_number==410000 && gen_met>200) return;
  if(channel_number==407012 && (gen_met<200 || gen_met>300)) return;
  if(channel_number==407322 && (gen_met<300 || gen_met>400)) return;
  if(channel_number==407323 && gen_met<400) return;

  // For single top. Below are currently Ht-filtered - this might need to be changed
  if(channel_number==410013 && gen_ht>500) return;
  if(channel_number==410014 && gen_ht>500) return;
  

  // baseline electrons are requested to pass the loose likelihood identification criteria
  //    and have pT > 5 GeV and |eta| < 2.47, with lose electron selections. For
  //    the lepton veto, pT > 20 GeV (below) is required
  auto electrons  = event->getElectrons(5, 2.47, ELooseLH);
  // baseline muons are required to pass the Medium selections and to have pT > 5 GeV, |eta| < 2.5
  auto muons      = event->getMuons(5, 2.5, MuMedium);
  // small-R jets: pT > 25 GeV, |eta| < 2.8
  auto candJets   = event->getJets(25., 2.8);
  // TODO: determine if MET includes TST at all. Doesn't look like it.
  auto metVec     = event->getMET();
  double met      = metVec.Et();
  auto ht 	  = sumObjectsPt(candJets); 
  double met_sig  = met/sqrt(ht);

  if(countObjects(candJets, 20, 2.8, NOT(LooseBadJet))!=0) return;
  // No bad muon veto implemented

  // TODO: do we need this? Let's leave it in here as this was here before
  candJets = filterObjects(candJets, 20, 2.8, JVT50Jet);
  // reclusterJets(jets, ReclusterRadius, RCJetPtMin, RCJetSubjetRadius, RCJetPtFrac)
  auto fatJets = reclusterJets(candJets, 0.8, 100, 0.2, 0.10);

  // Overlap removal
  // TODO: double-check that Overlap Removal defined correctly here. I looked at it and it appears to match INT note definitions
  auto radiusCalcJet  = [] (const AnalysisObject& , const AnalysisObject& muon) { return std::min(0.4, 0.04 + 10/muon.Pt()); };
  auto radiusCalcMuon = [] (const AnalysisObject& muon, const AnalysisObject& ) { return std::min(0.4, 0.04 + 10/muon.Pt()); };
  auto radiusCalcElec = [] (const AnalysisObject& elec, const AnalysisObject& ) { return std::min(0.4, 0.04 + 10/elec.Pt()); };

  electrons  = overlapRemoval(electrons, muons, 0.01);
  candJets   = overlapRemoval(candJets, electrons, 0.2, NOT(BTag77MV2c10));
  electrons  = overlapRemoval(electrons, candJets, radiusCalcElec);
  candJets   = overlapRemoval(candJets, muons, radiusCalcJet, LessThan3Tracks);
  muons      = overlapRemoval(muons, candJets, radiusCalcMuon);

  // No cosmic muon veto implemented

  // require signal jets to be 30 GeV
  auto signalJets      = filterObjects(candJets, 25);
  // signal electrons are required to pass the Medium likelihood criteria and isolated using LooseTrackOnly
  auto signalElectrons = filterObjects(electrons, 20, 2.47, ETightLH|ED0Sigma5|EZ05mm|EIsoBoosted);
  // signal muons are required to be isolated using LooseTrackOnly
  auto signalMuons     = filterObjects(muons, 20, 2.5, MuD0Sigma3|MuZ05mm|MuIsoBoosted|MuNotCosmic);
  // combine into signalLeptons for easy counting
  auto signalLeptons   = signalElectrons + signalMuons;

  // get b-jets
  auto candBJets = filterObjects(signalJets, 25., 2.5);
  auto bjets = filterObjects(signalJets, 25., 2.5, BTag77MV2c10);

  int n_bjets       = bjets.size();
  int n_jets        = signalJets.size();
  int n_leptons     = signalLeptons.size();

// Implement the truth-tagging for the TRF variables 
#ifdef ROOTCORE_PACKAGE_BTaggingTruthTagging
  // Need b-tagging weights
  std::vector<double>  m_TTweight_in;
  m_TTweight_in.clear();
  std::vector<double>  m_TTweight_ex;
  m_TTweight_ex.clear();
  
  // Need b-tagging permutations
  std::vector<std::vector<bool> > m_TTPerm_in;
  m_TTPerm_in.clear(); 
  std::vector<std::vector<bool> > m_TTPerm_ex;
  m_TTPerm_ex.clear(); 

  StatusCode code = m_btt->setSeed(n_jets + 10*n_leptons + 100*met + 1000*electrons.size() + 10000*muons.size());
  if (!code.isSuccess()) throw std::runtime_error("error in setting the seed");

  std::vector<double> pt   = std::vector<double>(candBJets.size(), 0);
  std::vector<double> eta  = std::vector<double>(candBJets.size(), 0);
  std::vector<int>    flav = std::vector<int>   (candBJets.size(), 1);
  std::vector<double> tagw = std::vector<double>(candBJets.size(), 1);

  for (unsigned int index = 0; index < candBJets.size(); ++index){
    auto jet = candBJets.at(index);
    pt[index]   = jet.Pt()*1000;
    eta[index]  = jet.Eta();
    flav[index] = jetFlavor(jet);
  }

  code = m_btt->setJets(&pt, &eta, &flav, &tagw);
  if (!code.isSuccess()) throw std::runtime_error("error setting the jets for truth tagging in execute()");
  code = m_btt->getTruthTagWei(4, m_TTweight_ex, m_TTweight_in);
  if (!code.isSuccess()) throw std::runtime_error("error in retrieving the weights");
  code = m_btt->getTagPermutation(4, m_TTPerm_ex, m_TTPerm_in); 
  if (!code.isSuccess()) throw std::runtime_error("error in retrieving the permutations");

#else
  // require at least 2 b-jets
  if(n_bjets<2) return;
#endif
  
  // *** Calculate the variables unaffected by TRF and non-TRF *** //
  // min mT of leading 3 bjets
  float mTb_min    = calcMTmin(bjets, metVec,3);
  // dphimin between leading 4 signal jets and met
  float dphiMin4  = minDphi(metVec, signalJets, 4);
  // leading lepton and met
  float mT = (signalLeptons.size()>0)? calcMT(signalLeptons[0], metVec) : 0.0;
  // sum of leading 4 reclustered jet masses
  float mjsum = sumObjectsM(fatJets, 4);
  // dPhi(j1, MET) for Gbb
  float dphi1jet = minDphi(metVec, signalJets, 1);
  // MET trigger - for truth-level studies just assume everything passes this
  
  if(n_jets>3) {

    // *** Perform the calculations using TRF here *** //
    std::vector<double> jetsForMeff_ex3b;
    jetsForMeff_ex3b.clear();
    std::vector<AnalysisObject> recordMeffJets_ex3b;
    recordMeffJets_ex3b.clear();
    std::vector<double> jetsForMeff_in4b;
    jetsForMeff_in4b.clear();
    std::vector<AnalysisObject> recordMeffJets_in4b_temp;
    recordMeffJets_in4b_temp.clear();
    std::vector<double> jetsForMeff_in3b;
    jetsForMeff_in3b.clear();
    std::vector<AnalysisObject> recordMeffJets_in3b;
    recordMeffJets_in3b.clear();
    std::vector<double> jetsForMeff_in2b;
    jetsForMeff_in2b.clear();
    std::vector<AnalysisObject> recordMeffJets_in2b;
    recordMeffJets_in2b.clear();
    std::vector<double> pt_tempVec_untagged;
    pt_tempVec_untagged.clear();
    std::vector<AnalysisObject> jet_tempVec_untagged;
    jet_tempVec_untagged.clear();
  
    // Store untagged jets
    double untagged_leadingPt = 0;
    int untaggedJetCounter = -1;
    for (const auto ijet : signalJets) {
      untaggedJetCounter++;
      pt_tempVec_untagged.push_back(ijet.Pt()*1000);
    }
    std::sort(pt_tempVec_untagged.begin(), pt_tempVec_untagged.end(), std::greater<double>());
    untagged_leadingPt = pt_tempVec_untagged.at(0);

    // Some counters for bookeeping 
    int ex3b = 0;
    // Store the == 3b-, >=3b-, and >= 4b-tagged jets
    int taggedJetCounter = -1;
    for (const auto ijet : candBJets) {
      taggedJetCounter++;
      if (candBJets.size() >= 3 && m_TTPerm_ex.at(3).size() != 0 && m_TTPerm_ex.at(3).at(taggedJetCounter) == 1) {
        ex3b++;
        jetsForMeff_ex3b.push_back(ijet.Pt()*1000);
        recordMeffJets_ex3b.push_back(ijet);
      }
    
      if (candBJets.size() >= 4 && m_TTPerm_in.at(4).size() != 0 && m_TTPerm_in.at(4).at(taggedJetCounter) == 1) {
        jetsForMeff_in4b.push_back(ijet.Pt()*1000);
        recordMeffJets_in4b_temp.push_back(ijet);
      }
    
      if (candBJets.size() >= 4 && m_TTPerm_in.at(3).size() != 0 && m_TTPerm_in.at(3).at(taggedJetCounter) == 1) {
        jetsForMeff_in3b.push_back(ijet.Pt()*1000);
        recordMeffJets_in3b.push_back(ijet);
      }
      
      if (candBJets.size() >= 4 && m_TTPerm_in.at(2).size() != 0 && m_TTPerm_in.at(2).at(taggedJetCounter) == 1) {
        jetsForMeff_in2b.push_back(ijet.Pt()*1000);
        recordMeffJets_in2b.push_back(ijet);
      }
    }
  
    for (int i = 0; i < pt_tempVec_untagged.size(); i++) {
      if (jetsForMeff_in3b.size() != 0 &&  jetsForMeff_ex3b.size() != 0 && ((pt_tempVec_untagged.at(i) != jetsForMeff_ex3b.at(0)) || (pt_tempVec_untagged.at(i) != jetsForMeff_ex3b.at(1)) || (pt_tempVec_untagged.at(i) != jetsForMeff_ex3b.at(2)))) {
        jetsForMeff_ex3b.push_back(pt_tempVec_untagged.at(i));
        for (const auto ijet : signalJets) {
          if (ijet.Pt()*1000 == pt_tempVec_untagged.at(i)) {
            recordMeffJets_ex3b.push_back(ijet);
            break;
          }
          else continue;
        }
        break;
      }
      else continue;
    } 
     

	
    for (int i = 0; i < pt_tempVec_untagged.size(); i++) {
      if (jetsForMeff_in3b.size() != 0 &&  jetsForMeff_ex3b.size() != 0 && ((pt_tempVec_untagged.at(i) != jetsForMeff_in3b.at(0)) || (pt_tempVec_untagged.at(i) != jetsForMeff_in3b.at(1)) || (pt_tempVec_untagged.at(i) != jetsForMeff_in3b.at(2)))) {
       jetsForMeff_in3b.push_back(pt_tempVec_untagged.at(i));
        for (const auto ijet : signalJets) {
          if (ijet.Pt()*1000 == pt_tempVec_untagged.at(i)) {
            recordMeffJets_in3b.push_back(ijet);
            break;
          }
          else continue;
        }
        break;
      }
      else continue;
    } 

    for (int i = 0; i < pt_tempVec_untagged.size(); i++) {
      if (jetsForMeff_in2b.size() != 0 && ((pt_tempVec_untagged.at(i) != jetsForMeff_in2b.at(0)) || (pt_tempVec_untagged.at(i) != jetsForMeff_in3b.at(1)) )) {
       jetsForMeff_in2b.push_back(pt_tempVec_untagged.at(i));
        for (const auto ijet : signalJets) {
          if (ijet.Pt()*1000 == pt_tempVec_untagged.at(i)) {
            recordMeffJets_in2b.push_back(ijet);
            break;
          }
          else continue;
        }
        break;
      }
      else continue;
    }
 
    std::vector<AnalysisObject> recordMeffJets_in4b;
    recordMeffJets_in4b.clear();
    if (candBJets.size() >= 4 && m_TTPerm_in.at(4).size() != 0 && m_TTPerm_in.at(4).at(taggedJetCounter) == 1) { 
      std::sort(jetsForMeff_in4b.begin(), jetsForMeff_in4b.end(), std::greater<double>());
      // Keep the 4 leading pT jets in the >= 4b jet vector
      jetsForMeff_in4b.resize(4); // Want the first four elements
      for (const auto ijet : recordMeffJets_in4b_temp) {
        if (ijet.Pt()*1000 == jetsForMeff_in4b.at(0)) recordMeffJets_in4b.push_back(ijet);
        if (ijet.Pt()*1000 == jetsForMeff_in4b.at(1)) recordMeffJets_in4b.push_back(ijet);
        if (ijet.Pt()*1000 == jetsForMeff_in4b.at(2)) recordMeffJets_in4b.push_back(ijet);
        if (ijet.Pt()*1000 == jetsForMeff_in4b.at(3)) recordMeffJets_in4b.push_back(ijet);    
      }
    }
  
    float sumJetsForMeff_ex3b = 0; 
    if (jetsForMeff_ex3b.size() != 0) {
      for (int jj = 0; jj < jetsForMeff_ex3b.size(); jj++) {
        sumJetsForMeff_ex3b += jetsForMeff_ex3b.at(jj);
      }
    }
    sumJetsForMeff_ex3b = sumJetsForMeff_ex3b/1000;
  
    float sumJetsForMeff_in3b = 0; 
    if (jetsForMeff_in3b.size() != 0) {
      for (int jj = 0; jj < jetsForMeff_in3b.size(); jj++) {
        sumJetsForMeff_in3b += jetsForMeff_in3b.at(jj);
      }
    }
    sumJetsForMeff_in3b = sumJetsForMeff_in3b/1000;
   
    float sumJetsForMeff_in4b = 0; 
    if (jetsForMeff_in4b.size() != 0) {
      for (int jj = 0; jj < jetsForMeff_in4b.size(); jj++) {
        sumJetsForMeff_in4b += jetsForMeff_in4b.at(jj);
      }
    }
    sumJetsForMeff_in4b = sumJetsForMeff_in4b/1000;
  
    float sumJetsForMeff_in2b = 0; 
    if (jetsForMeff_in2b.size() != 0) {
      for (int jj = 0; jj < jetsForMeff_in2b.size(); jj++) {
        sumJetsForMeff_in2b += jetsForMeff_in2b.at(jj);
      }
    }
    sumJetsForMeff_in2b = sumJetsForMeff_in2b/1000;
    
    // Calculation of the basic variables to cut on
    // inclusive - all jets, but leptons are not included 
    float meff_4bj_ex3b  = sumJetsForMeff_ex3b + met; 
    float meff_4bj_in3b  = sumJetsForMeff_in3b + met; 
    float meff_4bj_in4b  = sumJetsForMeff_in4b + met; 
    float meff_4bj_in2b  = sumJetsForMeff_in2b + met; 

    // Define the cuts required for the different signal, control, and validation regions
    // Different channels have slightly different preselections 
    float m1_ex3b = 0;
    float m2_ex3b = 0;
    float DeltaR_h1_dR_ex3b = 0;
    float DeltaR_h2_dR_ex3b = 0;
    float DeltaR_max_ex3b = 0;
  
    float dR1_ex3b = 999;
    float dR2_ex3b = 999;
    float dRmin_ex3b = 999;

    float m1_temp_ex3b = -999;
    float m2_temp_ex3b = -999;
    float dR1_temp_ex3b = -999;
    float dR2_temp_ex3b = -999;

    float m1_in3b = 0;
    float m2_in3b = 0;
    float DeltaR_h1_dR_in3b = 0;
    float DeltaR_h2_dR_in3b = 0;
    float DeltaR_max_in3b = 0;
  
    float dR1_in3b = 999;
    float dR2_in3b = 999;
    float dRmin_in3b = 999;

    float m1_temp_in3b = -999;
    float m2_temp_in3b = -999;
    float dR1_temp_in3b = -999;
    float dR2_temp_in3b = -999;

    float m1_in4b = 0;
    float m2_in4b = 0;
    float DeltaR_h1_dR_in4b = 0;
    float DeltaR_h2_dR_in4b = 0;
    float DeltaR_max_in4b = 0;
  
    float dR1_in4b = 999;
    float dR2_in4b = 999;
    float dRmin_in4b = 999;

    float m1_temp_in4b = -999;
    float m2_temp_in4b = -999;
    float dR1_temp_in4b = -999;
    float dR2_temp_in4b = -999;

    float m1_in2b = 0;
    float m2_in2b = 0;
    float DeltaR_h1_dR_in2b = 0;
    float DeltaR_h2_dR_in2b = 0;
    float DeltaR_max_in2b = 0;
  
    float dR1_in2b = 999;
    float dR2_in2b = 999;
    float dRmin_in2b = 999;

    float m1_temp_in2b = -999;
    float m2_temp_in2b = -999;
    float dR1_temp_in2b = -999;
    float dR2_temp_in2b = -999;
    
    int jjj = 0;
    std::vector<TLorentzVector> tlv_j_ex3b;
    tlv_j_ex3b.clear();
    std::vector<TLorentzVector> tlv_j_in3b;
    tlv_j_in3b.clear();
    std::vector<TLorentzVector> tlv_j_in4b;
    tlv_j_in4b.clear();
    std::vector<TLorentzVector> tlv_j_in2b;
    tlv_j_in2b.clear();
    if (recordMeffJets_ex3b.size() != 0) {
      for(const auto rjet : recordMeffJets_ex3b) {
        if (jjj > 3) break;
        TLorentzVector vj;
        vj.SetPtEtaPhiM( rjet.Pt(), rjet.Eta(), rjet.Phi(), rjet.M());
        jjj++;
        tlv_j_ex3b.push_back(vj);
      }
    }
    jjj = 0;
    if (recordMeffJets_in3b.size() != 0) {
      for(const auto rjet : recordMeffJets_in3b) {
        if (jjj > 3) break;
        TLorentzVector vj;
        vj.SetPtEtaPhiM( rjet.Pt(), rjet.Eta(), rjet.Phi(), rjet.M());
        jjj++;
        tlv_j_in3b.push_back(vj);
      }
    }
    jjj = 0;
    if (recordMeffJets_in4b.size() != 0) {
      for(const auto rjet : recordMeffJets_in4b) {
        if (jjj > 3) break;
        TLorentzVector vj;
        vj.SetPtEtaPhiM( rjet.Pt(), rjet.Eta(), rjet.Phi(), rjet.M());
        jjj++;
        tlv_j_in4b.push_back(vj);
      }
    }
    jjj = 0;
    if (recordMeffJets_in2b.size() != 0) {
      for(const auto rjet : recordMeffJets_in2b) {
        if (jjj > 3) break;
        TLorentzVector vj;
        vj.SetPtEtaPhiM( rjet.Pt(), rjet.Eta(), rjet.Phi(), rjet.M());
        jjj++;
        tlv_j_in2b.push_back(vj);
      }
    }
  

    TLorentzVector vjj1_ex3b;
    TLorentzVector vjj2_ex3b;
    TLorentzVector vjj1_in3b;
    TLorentzVector vjj2_in3b;
    TLorentzVector vjj1_in4b;
    TLorentzVector vjj2_in4b;
    TLorentzVector vjj1_in2b;
    TLorentzVector vjj2_in2b;
    std::vector<int> index_appo = {1,2,3};
    if (tlv_j_ex3b.size() == 4) {
      for (unsigned int i = 1; i< tlv_j_ex3b.size(); i++) {
        vjj1_ex3b = tlv_j_ex3b.at(0) + tlv_j_ex3b.at(i);
        float dR1_appo = tlv_j_ex3b.at(0).DeltaR(tlv_j_ex3b.at(i));
        std::vector<int> index = index_appo;
        index.erase(index.begin()+(i-1));
        vjj2_ex3b = tlv_j_ex3b.at(index.at(0)) + tlv_j_ex3b.at(index.at(1));
        float dR2_appo = tlv_j_ex3b.at(index.at(0)).DeltaR( tlv_j_ex3b.at(index.at(1))   );
        float m1_appo = vjj1_ex3b.M();
        float m2_appo = vjj2_ex3b.M(); 
        if(std::max(dR1_appo, dR2_appo) < dRmin_ex3b){
          dRmin_ex3b = std::max(dR1_appo, dR2_appo);
          m1_ex3b = m1_appo;
          m2_ex3b = m2_appo;
          dR1_ex3b = dR1_appo;
          dR2_ex3b = dR2_appo;
        }                                                      
      }
    }
    if (m1_ex3b > m2_ex3b) {
      m1_temp_ex3b = m1_ex3b;
      m2_temp_ex3b = m2_ex3b;
      dR1_temp_ex3b = dR1_ex3b;
      dR2_temp_ex3b = dR2_ex3b;  
    }
    else {
      m1_temp_ex3b = m2_ex3b;
      m2_temp_ex3b = m1_ex3b;
      dR1_temp_ex3b = dR2_ex3b;
      dR2_temp_ex3b = dR1_ex3b;  
    }  
  
    if (tlv_j_in3b.size() == 4) {
      for (unsigned int i = 1; i< tlv_j_in3b.size(); i++) {
        vjj1_in3b = tlv_j_in3b.at(0) + tlv_j_in3b.at(i);
        float dR1_appo = tlv_j_in3b.at(0).DeltaR(tlv_j_in3b.at(i));
        std::vector<int> index = index_appo;
        index.erase(index.begin()+(i-1));
        vjj2_in3b = tlv_j_in3b.at(index.at(0)) + tlv_j_in3b.at(index.at(1));
        float dR2_appo = tlv_j_in3b.at(index.at(0)).DeltaR( tlv_j_in3b.at(index.at(1))   );
        float m1_appo = vjj1_in3b.M();
        float m2_appo = vjj2_in3b.M(); 
        if(std::max(dR1_appo, dR2_appo) < dRmin_in3b){
          dRmin_in3b = std::max(dR1_appo, dR2_appo);
          m1_in3b = m1_appo;
          m2_in3b = m2_appo;
          dR1_in3b = dR1_appo;
          dR2_in3b = dR2_appo;
        }
                                                       
      }
    }
    if (m1_in3b > m2_in3b) {
      m1_temp_in3b = m1_in3b;
      m2_temp_in3b = m2_in3b;
      dR1_temp_in3b = dR1_in3b;
      dR2_temp_in3b = dR2_in3b;  
    }
    else {
      m1_temp_in3b = m2_in3b;
      m2_temp_in3b = m1_in3b;
      dR1_temp_in3b = dR2_in3b;
      dR2_temp_in3b = dR1_in3b;  
    } 
  
    if (tlv_j_in4b.size() == 4) {
      for (unsigned int i = 1; i< tlv_j_in4b.size(); i++) {
        vjj1_in4b = tlv_j_in4b.at(0) + tlv_j_in4b.at(i);
        float dR1_appo = tlv_j_in4b.at(0).DeltaR(tlv_j_in4b.at(i));
        std::vector<int> index = index_appo;
        index.erase(index.begin()+(i-1));
        vjj2_in4b = tlv_j_in4b.at(index.at(0)) + tlv_j_in4b.at(index.at(1));
        float dR2_appo = tlv_j_in4b.at(index.at(0)).DeltaR( tlv_j_in4b.at(index.at(1))   );
        float m1_appo = vjj1_in4b.M();
        float m2_appo = vjj2_in4b.M(); 
        if(std::max(dR1_appo, dR2_appo) < dRmin_in4b){
          dRmin_in4b = std::max(dR1_appo, dR2_appo);
          m1_in4b = m1_appo;
          m2_in4b = m2_appo;
          dR1_in4b = dR1_appo;
          dR2_in4b = dR2_appo;
        }
      }
    }
    if (m1_in4b > m2_in4b) {
      m1_temp_in4b = m1_in4b;
      m2_temp_in4b = m2_in4b;
      dR1_temp_in4b = dR1_in4b;
      dR2_temp_in4b = dR2_in4b;  
    }
    else {
      m1_temp_in4b = m2_in4b;
      m2_temp_in4b = m1_in4b;
      dR1_temp_in4b = dR2_in4b;
      dR2_temp_in4b = dR1_in4b;  
    }

    if (tlv_j_in2b.size() == 4) {
      for (unsigned int i = 1; i< tlv_j_in2b.size(); i++) {
        vjj1_in2b = tlv_j_in2b.at(0) + tlv_j_in2b.at(i);
        float dR1_appo = tlv_j_in2b.at(0).DeltaR(tlv_j_in2b.at(i));
        std::vector<int> index = index_appo;
        index.erase(index.begin()+(i-1));
        vjj2_in2b = tlv_j_in2b.at(index.at(0)) + tlv_j_in2b.at(index.at(1));
        float dR2_appo = tlv_j_in2b.at(index.at(0)).DeltaR( tlv_j_in2b.at(index.at(1))   );
        float m1_appo = vjj1_in2b.M();
        float m2_appo = vjj2_in2b.M(); 
        if(std::max(dR1_appo, dR2_appo) < dRmin_in2b){
          dRmin_in2b = std::max(dR1_appo, dR2_appo);
          m1_in2b = m1_appo;
          m2_in2b = m2_appo;
          dR1_in2b = dR1_appo;
          dR2_in2b = dR2_appo;
        }
                                                       
      }
    }
    if (m1_in2b > m2_in2b) {
      m1_temp_in2b = m1_in2b;
      m2_temp_in2b = m2_in2b;
      dR1_temp_in2b = dR1_in2b;
      dR2_temp_in2b = dR2_in2b;  
    }
    else {
      m1_temp_in2b = m2_in2b;
      m2_temp_in2b = m1_in2b;
      dR1_temp_in2b = dR2_in2b;
      dR2_temp_in2b = dR1_in2b;  
    } 

    // Rename the final variables for ease of use
    m1_ex3b = m1_temp_ex3b;
    m2_ex3b = m2_temp_ex3b;
    DeltaR_h1_dR_ex3b = dR1_temp_ex3b; 
    DeltaR_h2_dR_ex3b = dR2_temp_ex3b;
    DeltaR_max_ex3b = std::max(DeltaR_h1_dR_ex3b,DeltaR_h2_dR_ex3b);

    m1_in3b = m1_temp_in3b;
    m2_in3b = m2_temp_in3b;
    DeltaR_h1_dR_in3b = dR1_temp_in3b; 
    DeltaR_h2_dR_in3b = dR2_temp_in3b;
    DeltaR_max_in3b = std::max(DeltaR_h1_dR_in3b,DeltaR_h2_dR_in3b);
  
    m1_in4b = m1_temp_in4b;
    m2_in4b = m2_temp_in4b;
    DeltaR_h1_dR_in4b = dR1_temp_in4b; 
    DeltaR_h2_dR_in4b = dR2_temp_in4b;
    DeltaR_max_in4b = std::max(DeltaR_h1_dR_in4b,DeltaR_h2_dR_in4b);

    m1_in2b = m1_temp_in2b;
    m2_in2b = m2_temp_in2b;
    DeltaR_h1_dR_in2b = dR1_temp_in2b; 
    DeltaR_h2_dR_in2b = dR2_temp_in2b;
    DeltaR_max_in2b = std::max(DeltaR_h1_dR_in2b,DeltaR_h2_dR_in2b);
    
    // *** End of the calculations using TRF *** //

    // Fill with the TRF variables 
    fill("meff_4bj_ex3b", meff_4bj_ex3b);
    fill("meff_4bj_in3b", meff_4bj_in3b);
    fill("meff_4bj_in4b", meff_4bj_in4b);
    fill("meff_4bj_in2b", meff_4bj_in2b);
    fill("m1_ex3b", m1_ex3b);
    fill("m2_ex3b", m2_ex3b);
    fill("m1_in3b", m1_in3b);
    fill("m2_in3b", m2_in3b);
    fill("m1_in4b", m1_in4b);
    fill("m2_in4b", m2_in4b);
    fill("m1_in2b", m1_in2b);
    fill("m2_in2b", m2_in2b);
    fill("DeltaR_max_ex3b", DeltaR_max_ex3b);
    fill("DeltaR_max_in3b", DeltaR_max_in3b);
    fill("DeltaR_max_in4b", DeltaR_max_in4b);
    fill("DeltaR_max_in2b", DeltaR_max_in2b);
  
    // Add TRF variables to the n-tuple
    ntupVar("meff_4bj_ex3b", meff_4bj_ex3b);
    ntupVar("meff_4bj_in3b", meff_4bj_in3b);
    ntupVar("meff_4bj_in4b", meff_4bj_in4b);
    ntupVar("meff_4bj_in2b", meff_4bj_in2b);
    ntupVar("m1_ex3b", m1_ex3b);
    ntupVar("m2_ex3b", m2_ex3b);
    ntupVar("m1_in3b", m1_in3b);
    ntupVar("m2_in3b", m2_in3b);
    ntupVar("m1_in4b", m1_in4b);
    ntupVar("m2_in4b", m2_in4b);
    ntupVar("m1_in2b", m1_in2b);
    ntupVar("m2_in2b", m2_in2b);
    ntupVar("DeltaR_max_ex3b", DeltaR_max_ex3b);
    ntupVar("DeltaR_max_in3b", DeltaR_max_in3b);
    ntupVar("DeltaR_max_in4b", DeltaR_max_in4b);
    ntupVar("DeltaR_max_in2b", DeltaR_max_in2b);
  #ifdef ROOTCORE_PACKAGE_BTaggingTruthTagging
    ntupVar("weight_3b_in", m_TTweight_in.at(3));
    ntupVar("weight_3b_ex", m_TTweight_ex.at(3));
    ntupVar("weight_4b_in", m_TTweight_in.at(4));
    ntupVar("weight_2b_in", m_TTweight_in.at(2));
  #endif 

  }

  if(n_bjets > 2) {
    // *** Perform the calculations using no TRF here *** //
    std::vector<double> jetsForMeff;
    jetsForMeff.clear();
  
    std::vector<AnalysisObject> recordMeffJets;
    recordMeffJets.clear();
    if (n_bjets == 4) {
      for (int i = 0; i < n_bjets; i++) {
        jetsForMeff.push_back(bjets.at(i).Pt()*1000);
        recordMeffJets.push_back(bjets.at(i));
      }
    }
    else if (n_bjets > 4) {
      std::vector<double> pt_bjetVec;
      pt_bjetVec.clear();
      for (int i = 0; i < n_bjets; i++) {
        pt_bjetVec.push_back(bjets.at(i).Pt()*1000);
      }
      std::sort(pt_bjetVec.begin(), pt_bjetVec.end(), std::greater<float>());
      for (const auto ibjet : bjets) { if (ibjet.Pt()*1000 == pt_bjetVec.at(0)) recordMeffJets.push_back(ibjet); }
      for (const auto ibjet : bjets) { if (ibjet.Pt()*1000 == pt_bjetVec.at(1)) recordMeffJets.push_back(ibjet); }
      for (const auto ibjet : bjets) { if (ibjet.Pt()*1000 == pt_bjetVec.at(2)) recordMeffJets.push_back(ibjet); }
      for (const auto ibjet : bjets) { if (ibjet.Pt()*1000 == pt_bjetVec.at(3)) recordMeffJets.push_back(ibjet); }
      for (int ii = 0; ii < 4; ii++) {
        jetsForMeff.push_back(pt_bjetVec.at(ii));
      }
    }
    else {
      std::vector<double> pt_jetVec;
      pt_jetVec.clear();
      for (int j = 0; j < n_jets; j++) {
        pt_jetVec.push_back(signalJets.at(j).Pt()*1000);
      }
      std::sort(pt_jetVec.begin(), pt_jetVec.end(), std::greater<float>());
      if (n_bjets == 2) {
        for (const auto ijet : signalJets) { if (ijet.Pt()*1000 == pt_jetVec.at(0)) recordMeffJets.push_back(ijet); }
        for (const auto ijet : signalJets) { if (ijet.Pt()*1000 == pt_jetVec.at(1)) recordMeffJets.push_back(ijet); }
        recordMeffJets.push_back(bjets.at(0));
        recordMeffJets.push_back(bjets.at(1));
        jetsForMeff.push_back(pt_jetVec.at(0));
        jetsForMeff.push_back(pt_jetVec.at(1));
        jetsForMeff.push_back(bjets.at(0).Pt()*1000);
        jetsForMeff.push_back(bjets.at(1).Pt()*1000);
      }
      else {
        for (const auto ijet : signalJets) { if (ijet.Pt()*1000 == pt_jetVec.at(0)) recordMeffJets.push_back(ijet); }
        recordMeffJets.push_back(bjets.at(0)); 
        recordMeffJets.push_back(bjets.at(1)); 
        recordMeffJets.push_back(bjets.at(2)); 
        jetsForMeff.push_back(pt_jetVec.at(0));
        jetsForMeff.push_back(bjets.at(0).Pt()*1000);
        jetsForMeff.push_back(bjets.at(1).Pt()*1000);
        jetsForMeff.push_back(bjets.at(2).Pt()*1000);
      }
    }
    float sumPtJetsForMeff = 0;
    for (int jj = 0; jj < jetsForMeff.size(); jj++) {
      sumPtJetsForMeff += jetsForMeff.at(jj);
    }
    sumPtJetsForMeff = sumPtJetsForMeff/1000;
  
    // inclusive - all jets, but leptons are not included 
    // Need to now explicitly calculate the non-TRF variables 
    float meff_4bj  = sumPtJetsForMeff + met; 

    float m1 = 0;
    float m2 = 0;
    float DeltaR_h1_dR = 0;
    float DeltaR_h2_dR = 0;
    float DeltaR_max = 0;
  
    float dR1 = 999;
    float dR2 = 999;
    float dRmin = 999;

    float m1_temp = -999;
    float m2_temp = -999;
    float dR1_temp = -999;
    float dR2_temp = -999;

    int jjj = 0;
    std::vector<TLorentzVector> tlv_j;
    for(const auto rjet : recordMeffJets) {
      if (jjj > 3) break;
      TLorentzVector vj;
      vj.SetPtEtaPhiM( rjet.Pt(), rjet.Eta(), rjet.Phi(), rjet.M());
      jjj++;
      tlv_j.push_back(vj);
    }

    TLorentzVector vjj1;
    TLorentzVector vjj2;
    std::vector<int> index_appo = {1,2,3};
    for (unsigned int i = 1; i< tlv_j.size(); i++) {
      vjj1 = tlv_j.at(0) + tlv_j.at(i);
      float dR1_appo = tlv_j.at(0).DeltaR(tlv_j.at(i));
      std::vector<int> index = index_appo;
      index.erase(index.begin()+(i-1));
      vjj2 = tlv_j.at(index.at(0)) + tlv_j.at(index.at(1));
      float dR2_appo = tlv_j.at(index.at(0)).DeltaR( tlv_j.at(index.at(1))   );
      float m1_appo = vjj1.M();
      float m2_appo = vjj2.M();                                                        
      if (std::max(dR1_appo, dR2_appo) < dRmin) {
        dRmin = std::max(dR1_appo, dR2_appo);
        m1 = m1_appo;
        m2 = m2_appo;
        dR1 = dR1_appo;
        dR2 = dR2_appo;
      }
    }

    if (m1 > m2) {
      m1_temp = m1;
      m2_temp = m2;
      dR1_temp = dR1;
      dR2_temp = dR2;  
    }
    else {
      m1_temp = m2;
      m2_temp = m1;
      dR1_temp = dR2;
      dR2_temp = dR1;  
    }
  
    // Rename the final variables for ease of use
    m1 = m1_temp;
    m2 = m2_temp;
    DeltaR_h1_dR = dR1_temp; 
    DeltaR_h2_dR = dR2_temp;
    DeltaR_max = std::max(DeltaR_h1_dR, DeltaR_h2_dR);

    // Fill with the non-TRF variabes
    fill("meff_4bj", meff_4bj);
    fill("m1", m1);
    fill("m2", m2);
    fill("DeltaR_max", DeltaR_max);
    
    // Add the non-TRF variables to the n-tuple
    ntupVar("meff_4bj", meff_4bj);
    ntupVar("m1", m1);
    ntupVar("m2", m2);
    ntupVar("DeltaR_max", DeltaR_max);
  }
  // *** Fill the remaining variables, unaffected by TRF and non-TRF *** //
  fill("met", met);
  fill("met_sig", met_sig);
  fill("mjsum", mjsum);
  fill("mTb_min", mTb_min);
  fill("mT", mT);
  fill("dphiMin4", dphiMin4);
  fill("dphi1jet", dphi1jet);
  fill("n_jets", n_jets);
  fill("n_bjets", n_bjets);
  fill("n_electrons", signalElectrons.size());
  fill("n_muons", signalMuons.size());
  fill("n_leptons", n_leptons);
  fill("GenMET", gen_met);
  fill("GenHT", gen_ht);
  
  // *** Add the remaining variables, unaffected by TRF and non-TRF, to the n-tuple *** //
  ntupVar("mc_weight", event->getMCWeights()[0]);
  ntupVar("channel_number", channel_number);
  ntupVar("met", met);
  ntupVar("met_sig", met_sig);
  ntupVar("mjsum", mjsum);
  ntupVar("mTb_min", mTb_min);
  ntupVar("mT", mT);
  ntupVar("dphiMin4", dphiMin4);
  ntupVar("dphi1jet", dphi1jet);
  ntupVar("n_jets", n_jets);
  ntupVar("n_bjets", n_bjets);
  ntupVar("n_electrons", static_cast<int>(signalElectrons.size()));
  ntupVar("n_muons", static_cast<int>(signalMuons.size()));
  ntupVar("n_leptons", n_leptons);
  ntupVar("GenMET", gen_met);
  ntupVar("GenHT", gen_ht);
  
  return;
}
